# -*- coding: utf-8 -*-
#
#  File: accounts.py
#  Module: l10n_ch_vat_on_payment
#
#  Created by cyp@open-net.ch
#
#  Copyright (c) 2010 Open-Net Ltd. All rights reserved.
##############################################################################
#	
#	OpenERP, Open Source Management Solution
#	Copyright (C) 2004-2010 Tiny SPRL (<http://tiny.be>).
#
#	This program is free software: you can redistribute it and/or modify
#	it under the terms of the GNU Affero General Public License as
#	published by the Free Software Foundation, either version 3 of the
#	License, or (at your option) any later version.
#
#	This program is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU Affero General Public License for more details.
#
#	You should have received a copy of the GNU Affero General Public License
#	along with this program.  If not, see <http://www.gnu.org/licenses/>.	 
#
##############################################################################

import netsvc
from osv import fields, osv
import decimal_precision as dp

class account_account( osv.osv ):
	_inherit = 'account.account'

	def __compute(self, cr, uid, ids, field_names, arg=None, context=None, query='', query_params=()):
		""" compute the balance, debit and/or credit for the provided
		account ids
		Arguments:
		`ids`: account ids
		`field_names`: the fields to compute (a list of any of
					   'balance', 'debit' and 'credit')
		`arg`: unused fields.function stuff
		`query`: additional query filter (as a string)
		`query_params`: parameters for the provided query string
						(__compute will handle their escaping) as a
						tuple
		"""
		mapping = {
			'balance': "COALESCE(SUM(l.debit),0) " \
					   "- COALESCE(SUM(l.credit), 0) as balance",
			'debit': "COALESCE(SUM(l.debit), 0) as debit",
			'credit': "COALESCE(SUM(l.credit), 0) as credit"
		}
		#get all the necessary accounts
		children_and_consolidated = self._get_children_and_consol(cr, uid, ids, context=context)
		#compute for each account the balance/debit/credit from the move lines
		accounts = {}
		if children_and_consolidated:
			self.logger.notifyChannel('addons.'+self._name, netsvc.LOG_DEBUG, 'context is: %s'%str(context))
			ctx = context and context.copy() or {}
			ctx['handle_tva_state'] = context.get('handle_tva_state', False)
			aml_query = self.pool.get('account.move.line')._query_get(cr, uid, context=ctx)
			self.logger.notifyChannel('addons.'+self._name, netsvc.LOG_DEBUG, 'aml->query_get() gave: %s'%aml_query)

			wheres = [""]
			if query.strip():
				wheres.append(query.strip())
			if aml_query.strip():
				wheres.append(aml_query.strip())
			filters = " AND ".join(wheres)
			self.logger.notifyChannel('addons.'+self._name, netsvc.LOG_DEBUG, 'Filters: %s'%filters)
			# IN might not work ideally in case there are too many
			# children_and_consolidated, in that case join on a
			# values() e.g.:
			# SELECT l.account_id as id FROM account_move_line l
			# INNER JOIN (VALUES (id1), (id2), (id3), ...) AS tmp (id)
			# ON l.account_id = tmp.id
			# or make _get_children_and_consol return a query and join on that
			request = ("SELECT l.account_id as id, " +\
					   ', '.join(map(mapping.__getitem__, field_names)) +
					   " FROM account_move_line l" \
					   " WHERE l.account_id IN %s " \
							+ filters +
					   " GROUP BY l.account_id")
			params = (tuple(children_and_consolidated),) + query_params
			cr.execute(request, params)
			self.logger.notifyChannel('addons.'+self._name, netsvc.LOG_DEBUG,
									  'Status: %s'%cr.statusmessage)

			for res in cr.dictfetchall():
				accounts[res['id']] = res

			# consolidate accounts with direct children
			children_and_consolidated.reverse()
			brs = list(self.browse(cr, uid, children_and_consolidated, context=context))
			sums = {}
			currency_obj = self.pool.get('res.currency')
			while brs:
				current = brs[0]
#				can_compute = True
#				for child in current.child_id:
#					if child.id not in sums:
#						can_compute = False
#						try:
#							brs.insert(0, brs.pop(brs.index(child)))
#						except ValueError:
#							brs.insert(0, child)
#				if can_compute:
				brs.pop(0)
				for fn in field_names:
					sums.setdefault(current.id, {})[fn] = accounts.get(current.id, {}).get(fn, 0.0)
					for child in current.child_id:
						if child.company_id.currency_id.id == current.company_id.currency_id.id:
							sums[current.id][fn] += sums[child.id][fn]
						else:
							sums[current.id][fn] += currency_obj.compute(cr, uid, child.company_id.currency_id.id, current.company_id.currency_id.id, sums[child.id][fn], context=context)
			res = {}
			null_result = dict((fn, 0.0) for fn in field_names)
			for id in ids:
				res[id] = sums.get(id, null_result)
			return res

	_columns = {
		'balance': fields.function(__compute, digits_compute=dp.get_precision('Account'), method=True, string='Balance', multi='balance'),
		'credit': fields.function(__compute, digits_compute=dp.get_precision('Account'), method=True, string='Credit', multi='balance'),
		'debit': fields.function(__compute, digits_compute=dp.get_precision('Account'), method=True, string='Debit', multi='balance'),
	}

account_account()

class account_tax_code(osv.osv):
    """
    A code for the tax object.

    This code is used for some tax declarations.
    """
    def _sum(self, cr, uid, ids, name, args, context, where ='', where_params=()):
        parent_ids = tuple(self.search(cr, uid, [('parent_id', 'child_of', ids)]))
        if context.get('based_on', 'invoices') == 'payments':
            inv_state = 'paid'
            if context.get('handle_tva_state', False): inv_state = 'tva'
            print " ******* Here: inv_state=",inv_state
            cr.execute("""SELECT line.tax_code_id, sum(line.tax_amount)
                    FROM account_move_line AS line,
                        account_move AS move
                        LEFT JOIN account_invoice invoice ON
                            (invoice.move_id = move.id)
                    WHERE line.tax_code_id IN %s """+where+"""
                        AND move.id = line.move_id 
                        AND ((invoice.state = '%s')
                            OR (invoice.id IS NULL)) 
                            GROUP BY line.tax_code_id""" % inv_state,
                                (parent_ids,) + where_params)
        else:
            compl = " AND line.state='valid' "
            if context.get('handle_tva_state', False): compl = " AND line.state='tva' "
            print " ******* There: compl=",compl
            cr.execute("""SELECT line.tax_code_id, sum(line.tax_amount)
                    FROM account_move_line AS line,
                    account_move AS move
                    WHERE line.tax_code_id IN %s """+where+"""
                    AND move.id = line.move_id""" + compl + """
                    GROUP BY line.tax_code_id""",
                       (parent_ids,) + where_params)
        res = dict(cr.fetchall())
        res2 = {}
        obj_precision = self.pool.get('decimal.precision')
        for record in self.browse(cr, uid, ids, context=context):
            def _rec_get(record):
                amount = res.get(record.id, 0.0)
                for rec in record.child_ids:
                    amount += _rec_get(rec) * rec.sign
                return amount
            res2[record.id] = round(_rec_get(record), obj_precision.precision_get(cr, uid, 'Account'))
        return res2

    _inherit = 'account.tax.code'

account_tax_code()
